PVector xmin = new PVector( -100, -100 );
PVector xmax = new PVector( 100, 100 );
float isoV = 0.0;
int N = 16;

int currentMethod = 1; // 0: Hermite, 1: Chebyshev
String[] methodName = {"Hermite", "Chebyshev"};
boolean isSelected[] = { true, false, false, false };

//PVector uvs[] = { new PVector( 0.125, 0.0 ), new PVector( 1.0, 0.5 ), new PVector( 0.5, 1.0 ), new PVector( 0.0, 0.5 ) };
// the UV coordinates of the two endpoints of the edges; uvs0 is the endpoint that will correspond to the t=0 parameter, uvs1 the one for t=1
PVector uvs0[] = { new PVector( 0.0, 1.0 ), new PVector( 1.0, 0.0 ), new PVector( 0.0, 0.0 ), new PVector( 0.0, 0.0 ) };
PVector uvs1[] = { new PVector( 1.0, 1.0 ), new PVector( 1.0, 1.0 ), new PVector( 1.0, 0.0 ), new PVector( 0.0, 1.0 ) };

PVector bezierAABB[] = { new PVector( width/2, height ), new PVector( 0, 0 ) };

int approximationOrder = 1;

PlanarMapping   currentMapping    = new RadialHomographyMapping();
//ParametricCurve boundaries[]      = new TaylorCurve[4];
ParametricCurve boundaries[]; //      = new HermiteCurve[4];
BezierN         boundaryBeziers[] = new BezierN[4];

boolean doErrorBound = false;
boolean doAABB = true;
    
float mainOffsetX = 280;
float mainOffsetY = 280;
  
int boundaryCurveCount = 4;

void createHermites()
{
  boundaries = new HermiteCurve[4];
  // create the Hermite approximations
  for ( int i=0; i<boundaryCurveCount; ++i ) 
  {
    boundaries[i] = new HermiteCurve( currentMapping, approximationOrder, uvs0[i], uvs1[i] );
    boundaries[i].offsetX = width/2 ;
    boundaries[i].a = 0.0; //-c/2; //xmin.x;
    boundaries[i].b = 1.0; //c/2; //xmax.x;
    ((HermiteCurve)boundaries[i]).doErrorBound = doErrorBound;
  }
  // since Hermite curves are already in Bernstein basis, simply compute the AABB for them via BezierN then accumulate these AABBs into a global one called bezierAABB
  ((HermiteCurve)boundaries[0]).bez.updateAABB();
  bezierAABB[0].x = ((HermiteCurve)boundaries[0]).bez.aMin.x;
  bezierAABB[0].y = ((HermiteCurve)boundaries[0]).bez.aMin.y;
  bezierAABB[1].x = ((HermiteCurve)boundaries[0]).bez.bMax.x;
  bezierAABB[1].y = ((HermiteCurve)boundaries[0]).bez.bMax.y;
  for ( int i=1; i<boundaryCurveCount; ++i )
  {
    ((HermiteCurve)boundaries[i]).bez.updateAABB();
    bezierAABB[0].x = min( bezierAABB[0].x, ((HermiteCurve)boundaries[i]).bez.aMin.x );
    bezierAABB[0].y = min( bezierAABB[0].y, ((HermiteCurve)boundaries[i]).bez.aMin.y );
    bezierAABB[1].x = max( bezierAABB[1].x, ((HermiteCurve)boundaries[i]).bez.bMax.x );
    bezierAABB[1].y = max( bezierAABB[1].y, ((HermiteCurve)boundaries[i]).bez.bMax.y );    
  }  
}

// This approach is gradient-free so you don't have to provide derivatives, only bounds on the derivatives, if you need error expansion
void createChebyshevs()
{
  boundaries = new ChebyshevCurve[4];
  // create the Chebyshev node polynomial interpolations
  float w = xmax.x - xmin.x;
  float h = xmax.y - xmin.y;
  
  for ( int i=0; i<boundaryCurveCount; ++i ) 
  {
    boundaries[i] = new ChebyshevCurve( currentMapping, approximationOrder, uvs0[i], uvs1[i] );
    boundaries[i].offsetX = width/2 ;
    boundaries[i].a = 0.0; //-c/2; //xmin.x;
    boundaries[i].b = 1.0; //c/2; //xmax.x;
    ((ChebyshevCurve)boundaries[i]).doErrorBound = doErrorBound;
  }
  // since Hermite curves are already in Bernstein basis, simply compute the AABB for them via BezierN then accumulate these AABBs into a global one called bezierAABB
  ((ChebyshevCurve)boundaries[0]).bez.updateAABB();
  bezierAABB[0].x = ((ChebyshevCurve)boundaries[0]).bez.aMin.x;
  bezierAABB[0].y = ((ChebyshevCurve)boundaries[0]).bez.aMin.y;
  bezierAABB[1].x = ((ChebyshevCurve)boundaries[0]).bez.bMax.x;
  bezierAABB[1].y = ((ChebyshevCurve)boundaries[0]).bez.bMax.y;
  for ( int i=1; i<boundaryCurveCount; ++i )
  {
    ((ChebyshevCurve)boundaries[i]).bez.updateAABB();
    bezierAABB[0].x = min( bezierAABB[0].x, ((ChebyshevCurve)boundaries[i]).bez.aMin.x );
    bezierAABB[0].y = min( bezierAABB[0].y, ((ChebyshevCurve)boundaries[i]).bez.aMin.y );
    bezierAABB[1].x = max( bezierAABB[1].x, ((ChebyshevCurve)boundaries[i]).bez.bMax.x );
    bezierAABB[1].y = max( bezierAABB[1].y, ((ChebyshevCurve)boundaries[i]).bez.bMax.y );    
  }  
}

void dispatchMethod()
{
  switch ( currentMethod )
  {
    case 0:
      createHermites();
      break;
    case 1:
      createChebyshevs();
      break;
  }
}

void setup()
{
  size(1200, 600);
  textSize( 12 );
  
  dispatchMethod();
  
  smooth();
  rectMode(CORNERS);
}

PVector getPoint( float u, float v )
{
  float tx = (1-u) * xmin.x + u * xmax.x;
  float ty = (1-v) * xmin.x + v * xmax.y;
      
  return new PVector( tx, ty );  
}

void draw()
{
  background( 255, 255, 255 );
  
  //  dispatchMethod();
  
  pushMatrix();
  translate(mainOffsetX, mainOffsetY);
  
  // cell in the base plane
  strokeWeight( 4 );
  for (int i=0; i<=N; ++i)
  {
    for (int j=0; j<=N; ++j)
    {
      float u = i / float( N );
      float v = j / float( N );
      
      PVector pt = getPoint( u, v );
      
      stroke( u*255, 0, 0 );
      
      point( pt.x, pt.y );
    }
  }
  popMatrix();
  
  // v = isoV isoline on cell plane
  strokeWeight( 2 );
  stroke( 128 );
  beginShape();
  for (int i=0; i<=N; ++i)
  {
    float u = i / float( N );
    PVector pt = getPoint( u, isoV );
    vertex( pt.x, pt.y );
  }
  endShape();
  
  // splitter between the two planes
  strokeWeight( 2 );
  stroke( 255, 0, 0 );
  line( width/2, 0, width/2, height );
  
  pushMatrix();
  translate(mainOffsetX, mainOffsetY);
  
  // transformed image of the cell from plane 1 on plane 2
  strokeWeight( 4 );
  for (int i=0; i<=N; ++i)
  {
    for (int j=0; j<=N; ++j)
    {
      float u = i / float( N );
      float v = j / float( N );
      
      PVector preTransPt = getPoint( u, v );
      
      stroke(255*u, 0, 0);
      
      PVector pt = currentMapping.f( preTransPt );
            
      point( width/2 + pt.x, pt.y );
    }
  }
  
  // v = isoV isoline of transformed cell
  strokeWeight( 2 );
  stroke( 128 );
  noFill();
  beginShape();
  for (int i=0; i<4*N; ++i)
  {
    float u1 = i / float( 4*N );
    
    PVector preTransPt1 = getPoint( u1, isoV );
    PVector pt1 = currentMapping.f( preTransPt1 );

    vertex( width/2 + pt1.x, pt1.y );
  }
  endShape();
  
  pushMatrix();
  translate(width/2,0);
  for ( int i=0; i<boundaryCurveCount; ++i )
  {
    boundaries[i].draw();
  }
  popMatrix();
  
  // render AABB if the user wants to
  if ( doAABB )
  {
    noFill();
    strokeWeight(4);
    stroke(0,0,255, 64);
    rect( bezierAABB[0].x + width/2, bezierAABB[0].y, bezierAABB[1].x + width/2, bezierAABB[1].y);
  }
  popMatrix();
  
  // UI
  fill(0);
  stroke(0);
  text( "Plane 1", 20, 20);
  text( "Plane 2", width/2 + 20, 20 );

  text( "'m': switch method (Chebyshev, Hermite)", 20, 40);
  text( "'i'/'j': change order (degree = 2*order +1)", 20, 60);
  text( "'w'/'s': change lambda of radial homography", 20, 80);
  text( "'b': toggle AABB display", 20, 100);
  
  text( "Approximation order: " + approximationOrder, width/2 + 20, 60 );
  text( "Method: " + methodName[currentMethod], width/2 + 20, 80 );
  if ( doAABB )
  {
    float boundArea = ( bezierAABB[1].x - bezierAABB[0].x ) * ( bezierAABB[1].y - bezierAABB[0].y );
    text( "AABB bound area: " + boundArea, width/2 + 20, 100 );
  }
}

void keyPressed()
{
  int diffFromOne = key - '1';
  if ( diffFromOne >= 0 && diffFromOne < 4 ) isSelected[diffFromOne] = !isSelected[diffFromOne];
  if ( key == 'i' ) approximationOrder = ( approximationOrder + 1 ) % 4;
  if ( key == 'j' ) { approximationOrder = ( approximationOrder == 0 ) ? 3 : ( approximationOrder + 3 ) % 4;  }
  if ( key == 'e' )
  {
    doErrorBound = !doErrorBound;
  }
  if ( key == 'w' ) 
  {
    float lambda = ((RadialHomographyMapping)currentMapping).lambda + 0.25 / 65536.0;
    
    if (lambda < 1.05367435E-4)
      ((RadialHomographyMapping)currentMapping).lambda = lambda;
  }
  
  if ( key == 's' )
  {
    float lambda = ((RadialHomographyMapping)currentMapping).lambda - 0.25 / 65536.0;
    
    if (lambda > -2.4332272E-5)
      ((RadialHomographyMapping)currentMapping).lambda = lambda;
  }
  
  if ( key == 'm' ) currentMethod = ( currentMethod + 1 ) % 2;
  
  if ( key == 'b' ) doAABB = !doAABB;  
  
  dispatchMethod();
}
