class TaylorCurve extends ParametricCurve
{
  // uv \in [0,1]^2 is the position of the Taylor evaluation, for this particular cell
  TaylorCurve( PlanarMapping _mapping, int _n, PVector uv )
  {
    n = _n;
    mapping = _mapping;
    
    p0 = getPoint( uv.x, uv.y );
    
    p0.x -= xmin.x + (xmax.x - xmin.x) / 2.0;
    p0.y -= xmin.y + (xmax.y - xmin.y) / 2.0;
    
    q0 = mapping.f( p0 );
    
    sx = ( uv.x == 0.5 ) ? ( ( uv.y == 0.0 ) ? -1.0 : 1.0 ) : 0.0;
    sy = ( uv.y == 0.5 ) ? ( ( uv.x == 0.0 ) ? -1.0 : 1.0 ) : 0.0;      
    
    if ( uv.x == 0.5 ) isHorizontal = true;
    else               isHorizontal = false;

    df = new PVector[n];   
    for ( int i=0; i < n; ++i )
    {
      int dx = ( isHorizontal ) ? i+1 : 0;
      int dy = ( isHorizontal ) ? 0   : i+1;
      df[i] = mapping.df( dx, dy, p0 );
    }    
  }  
  
  PVector eval( float t )
  {
    return eval( q0, t );
  }
  
  //                   0! 1! 2! 3!  4!   5!   6!
  float factorial[] = { 1, 1, 2, 6, 24, 120, 720 };
  
  PVector eval( PVector pt, float t )
  {
    PVector result = pt;
    float tn = t;
    
    for ( int i=0; i<n; ++i)
    {
      //result = PVector.add( result, PVector.mult( df[i], pow( t, i+1 ) / factorial[i+1] ) );
      result = PVector.add( result, PVector.mult( df[i], tn ) );
      tn    *= t / float(i+2);
    }
    
    if ( doErrorBound )
    {
      PVector M = mapping.boundDf( n+1 );
      
      // careful: for horizontal lines we have to add the error bound _vertically_ and vice versa!
      M.y *= sx;
      M.x *= sy;
      
      // also careful: we need the absolute of tn, see the phrasing of the remainder and its bound using a derivative bound!
      result = PVector.add( result, PVector.mult( M, abs( tn ) ) );
    }
    
    return result;
  }
  
  PlanarMapping mapping; // the mapping that we approximate
  PVector df[]; // array of derivatives at the Taylor expansion point  
  int n; // degree
  float sx; // which direction to extend the boundary curve
  float sy; // which direction to extend the boundary curve
  PVector p0; // the evaluation position in Plane 1 (we do the Taylor expansion around this)
  PVector q0; // the evaluation position in Plane 2
  boolean doErrorBound; // whether to use error bounds
  boolean isHorizontal = false;
}
